.g-job-info-key
  font-weight bold
  margin 12px 8px 0 0
  font-size 16px

  &.inline
    display inline-block

.g-job-info-value.inline
  display inline-block

.g-job-status-badge
  border-radius 2px
  font-weight bold
  display inline
  padding 4px 9px 4px 4px
  text-transform uppercase
  font-size 13px

.g-job-cancel
  margin-left 10px
  padding 2px 8px

.g-job-timeline-container
  margin-top 5px
