.g-access-modal-body
  padding-top 10px
  padding-bottom 0

.g-public-container
  margin-bottom 16px

  .g-action-manage-public-flags
    margin-left 20px

  .radio
    margin-top 0
    color #aaa

  .radio.g-selected
    color #333

    i.icon-globe
      color #3355a0
    i.icon-lock
      color #d0b822

.g-ac-list-title
  background-color #888
  color white
  font-weight bold
  padding 3px 10px

i.g-flag-icon
  position relative

.g-flag-count-indicator
  display inline-block
  border-radius 3px
  background-color #c80000
  color white
  position absolute
  font-size 9px
  bottom -4px
  right -1px
  padding 0 0.3em
  font-style normal

.g-ac-list
  border 1px solid #d7d7d7
  border-top none
  background linear-gradient(to bottom, #d3d3d3 0, #f0f0f0 6px)
  padding 1px 4px

  .g-group-access-entry, .g-user-access-entry
    margin 9px 0

  .g-access-icon
    color #666
    font-size 25px
    position absolute

  .g-access-desc
    position relative
    left 42px
    width 310px
    .g-desc-title
      font-size 14px
    .g-desc-subtitle
      font-size 12px
      color #888

  .g-access-col-left
    display inline-block
    vertical-align top
    width 360px
    word-wrap break-word

  .g-access-col-right
    display inline-block
    vertical-align top

    .g-access-action-container
      margin-left 9px
      display inline-block

      a
        font-size 16px
        display inline-block
        vertical-align middle
        color #999
        margin-left 5px

      .g-action-manage-flags:hover
        color #0e84b5

      .g-action-remove-access:hover
        color #d02344

    select
      display inline-block
      width 113px

.g-grant-access-container
  margin-top 5px
  padding-top 8px

  .form-group>label
    font-weight normal
    font-size 13px
